/***********************************************************************
 WindowsStatusReporter.cpp - Defines the WindowsStatusReporter class.  
 	It provides some skeleton code for a generic Windows-based 
	StatusReporter derivative.

 $Revision: 1.1 $ $Date: 2001/10/06 12:01:58 $
	
 Copyright (c) 1996-2000 by ETR..., Inc.  See the file LICENSE which 
 accompanied this software for license details.
***********************************************************************/

#include "stdafx.h"

#include "WindowsStatusReporter.h"
#include "Headsman.h"

using namespace std;
using namespace DebugLevels;


////////////////////////////////////////////////////////////////////////
// Imports

// Pointer to the singleton -- also handles destroying the singleton
// when the program goes away.
extern Headsman<StatusReporter> gInstance_;


////////////////////////////////////////////////////////////////////////
// Constants

// Size of the string wrapping buffer -- if anything, this should be
// larger than the temp buffer and larger than the stream buffers
// defined by our parent class.
const int kWrappingBufferSize = 600;

// Width, in pixels, to wrap fatal error dialog messages to
const int kDefaultWrapWidth = 250;


//// GetInstance ///////////////////////////////////////////////////////
// Returns a pointer to the current instance of this Singleton, creating
// it if necessary.

StatusReporter* WindowsStatusReporter::GetInstance()
{
	if (gInstance_.Get() == 0) {
		gInstance_.Set(new WindowsStatusReporter());
	}
	return gInstance_.Get();
}


//// ReportStatus //////////////////////////////////////////////////////
// Handles the grunt work of loading status messages as string
// resources for subclasses.  It passes the handling of non-resource
// messages and actually displaying these messages off to the subclass.

StatusReporter::ErrorCode WindowsStatusReporter::ReportStatus(
		int nStatusCode, const char* pcFile, int nLine)
{
	// Give the superclass first chance to handle the status.  Really,
	// the superclass doesn't do much other than consult our subclass
	// through the GetStatusMessage template function.
	StatusReporter::ErrorCode eErrorCode = StatusReporter::ReportStatus(
			nStatusCode, pcFile, nLine);
	if (eErrorCode == StatusReporter::kReportFailed) {
		DebugLevel eDebugLevel = InferDebugLevel(nStatusCode);
		ostream& os = GetStream();

		// Our subclass has no special input, so try the default case:
		// loading the status message as a string resource.
		CString sTemp;
		if (sTemp.LoadString(nStatusCode)) {
			os << sTemp;
		}
		else {
			// If we get in here, some hoser probably passed us a bad
			// status code.  Ergo, the program's broken, and a "Duh, I
			// don't know" message is quite appropriate.
			os << "Error #" << GetLastError() << " while loading "
					"status message #" << nStatusCode << ". Please "
					"call technical support.";
			eDebugLevel = kFatalError;
		}

		// By now we have a textual status message of some sort.  So,
		// send it to the next handling stage.
		eErrorCode = ReportStatus(os, eDebugLevel, pcFile, nLine);
	}

    return eErrorCode;
}


//// ReportStatus //////////////////////////////////////////////////////
// Second override: the previous stages have built up the status string,
// so here we wrap the message if it's going to be sent to a message
// box.

StatusReporter::ErrorCode WindowsStatusReporter::ReportStatus(
		Postream& os, DebugLevel eDebugLevel, const char* pcFile,
		int nLine)
{
	// Let the base class handle some basics for us
	StatusReporter::ErrorCode e = StatusReporter::ReportStatus(os,
			eDebugLevel, pcFile, nLine);
	if (e == StatusReporter::kIgnoreReport) {
		// Someone has decided that we should ignore this report.  So,
		// just quietly log it rather than shout it out.
		os << " (ignored level " << eDebugLevel << " status)";
		eDebugLevel = kVerboseStatus;
	}
	else if (e == StatusReporter::kDemoteErrorLevel) {
		// Just another fatal error, so quietly demote its debug level
		// so we don't blast the user with streams of fatal errors.
		os << " (demoted fatal error)";
		eDebugLevel = kVerboseStatus;
	}

	// Almost ready to display it, so tack null on end, forcibly if need
	// be.
	const char* pcStatusMessage = GetBuffer(os);
	char* pcNewMessage = const_cast<char*>(pcStatusMessage);
	os << Pends;
	pcNewMessage[GetBufferSize() - 1] = '\0';

	if ((eDebugLevel == kFatalError) && (pcStatusMessage != 0) &&
				(pcStatusMessage[0] != '\0')) {
		// Add newlines to pcStatusMessage so it wraps properly.
		pcNewMessage = WrapMessage(pcStatusMessage);
	}

	DisplayStatusMessage(pcNewMessage, eDebugLevel, pcFile, nLine);
	ReleaseStream(os);

    return StatusReporter::kNoError;
}


//// GetStatusMessage //////////////////////////////////////////////////
// Just returns false to indicate the default behavior: no special
// status message loading performed.  We override StatusReporter's
// stupid version of this function because the default behavior is more
// useful to us.

bool WindowsStatusReporter::GetStatusMessage(ostream&, int)
{
	return false;
}


//// WrapMessage ///////////////////////////////////////////////////////
// Takes a pointer to a message string and adds newlines to it so that
// it wraps to nWidth pixels; it uses a standard screen DC with the
// default font for the calculation.  Returns a pointer to the new
// string.  The returned pointer points to a static buffer, so callers
// should copy the contents of this string before using it.

char* WindowsStatusReporter::WrapMessage(const char* pcMessage,
		int nWidth)
{
	CWindowDC dc(CWnd::FromHandle(0));              // 0 == whole screen
	static char acWrappingBuffer[kWrappingBufferSize];
	const char* pcLineStart = pcMessage;
	int nChars;

	if (nWidth == 0) {
		// A value wasn't given for the width, so just use a reasonable
		// default.
		nWidth = kDefaultWrapWidth;
	}

	// Zero first byte of wrapping buffer so first strcat works right --
	// have to do it here instead of as an initializer because wrapping
	// buffer is static.
	acWrappingBuffer[0] = '\0';

	while (*pcMessage != '\0') {
		nChars = 0;

		do {
			// Skip to next word
			while ((*pcMessage != '\0') && (*pcMessage != ' ')) {
				++nChars;
				++pcMessage;
			}

			// Breeze past any extra spaces
			while (*pcMessage == ' ') {
				++nChars;
				++pcMessage;
			}
		} while ((*pcMessage != '\0') &&
				(dc.GetTextExtent(pcLineStart, nChars).cx < nWidth));

		// Tack contents of line onto output string and terminate it
		// with a CRLF and a null.
		strncat(acWrappingBuffer, pcLineStart, nChars);
		if (*pcMessage != '\0') {
			strcat(acWrappingBuffer, "\r\n");
		}

		// Move line start pointer to beginning of next line.
		pcLineStart += nChars;
	} // end while (*pcMessage != '\0')

	return acWrappingBuffer;
}

